/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include "Midi/CommonMidiUtils.h"
#include "IO/MidiToMemoryStream.h"
#include "IO/IOUtils.h"

#include "Midi/Sequence.h"
#include "Midi/Track.h"
#include "Midi/Players/PlatformMidiManager.h"

#include "GUI/GLPane.h"

#include "jdkmidi/world.h"
#include "jdkmidi/track.h"
#include "jdkmidi/multitrack.h"
#include "jdkmidi/filereadmultitrack.h"
#include "jdkmidi/fileread.h"
#include "jdkmidi/fileshow.h"
#include "jdkmidi/filewritemultitrack.h"
#include "jdkmidi/msg.h"
#include "jdkmidi/sysex.h"

#include "jdkmidi/sequencer.h"
#include "jdkmidi/driver.h"
#include "jdkmidi/process.h"

#include <wx/timer.h>

#include <iostream>

namespace AriaMaestosa {
	
	/*
	 void	SetStatus( unsigned char s )	{ status=s;		}
	 void	SetChannel( unsigned char s )	{status=(unsigned char)((status&0xf0)|s);}
	 void	SetType( unsigned char s )		{status=(unsigned char)((status&0x0f)|s);}
	 
	 void	SetByte1( unsigned char b )		{ byte1=b;		}
	 void	SetByte2( unsigned char b )		{ byte2=b;		}
	 void	SetByte3( unsigned char b )		{ byte3=b;		}
	 
	 void	SetNote( unsigned char n ) 		{ byte1=n;		}
	 void	SetVelocity(unsigned char v) 	{ byte2=v;		}
	 void	SetPGValue(unsigned char v) 	{ byte1=v;		}
	 void	SetController(unsigned char c) 	{ byte1=c;		}
	 void	SetControllerValue(unsigned char v ) { byte2=v;		}
	 
	 void	SetBenderValue( short v);
	 void	SetMetaType( unsigned char t ) ;
	 void	SetMetaValue( unsigned short v );
	 void	SetNoteOn( unsigned char chan, unsigned char note, unsigned char vel );
	 void	SetNoteOff( unsigned char chan, unsigned char note, unsigned char vel );
	 void	SetPolyPressure( unsigned char chan, unsigned char note, unsigned char pres );
	 void	SetControlChange( unsigned char chan, unsigned char ctrl, unsigned char val );
	 void	SetProgramChange( unsigned char chan, unsigned char val );
	 void	SetChannelPressure( unsigned char chan, unsigned char val );
	 void	SetPitchBend( unsigned char chan, short val );
	 void	SetPitchBend( unsigned char chan, unsigned char low, unsigned char high );
	 void	SetSysEx();
	 void	SetMTC( unsigned char field, unsigned char v );
	 void	SetSongPosition( short pos );
	 void	SetSongSelect(unsigned char sng);
	 void	SetTuneRequest();
	 void	SetMetaEvent( unsigned char type, unsigned char v1, unsigned char v2 );
	 void	SetMetaEvent( unsigned char type, unsigned short v );
	 void	SetAllNotesOff( unsigned char chan, unsigned char type=C_ALL_NOTES_OFF );
	 void	SetLocal( unsigned char chan, unsigned char v );
	 void	SetNoOp();
	 void	SetTempo32( unsigned short tempo_times_32 );
	 void	SetText( unsigned short text_num, unsigned char type=META_GENERIC_TEXT );
	 void	SetDataEnd();
	 void	SetTimeSig( unsigned char numerator, unsigned char denominator );
	 void   SetKeySig( signed char sharp_flats, unsigned char major_minor );
	 void	SetBeatMarker();
	 
	 
	 */
	
bool exportMidiFile(Sequence* sequence, char* filepath)
{
		// when we're saving, we always want song to start at first measure, so temporarly switch firstMeasure to 0, and set it back in the end
		const int firstMeasureValue=sequence->firstMeasure;
		sequence->firstMeasure=0;   
		
		jdkmidi::MIDIMultiTrack tracks;
		int length = -1, start = -1, numTracks = -1;
		makeJDKMidiSequence(sequence, tracks, false, &length, &start, &numTracks);
		
		jdkmidi::MIDIFileWriteStreamFileName file_stream( filepath );
		
		jdkmidi::MIDIFileWriteMultiTrack writer2(
												 &tracks,
												 &file_stream
												 );
		
		// write the output file
		if( !writer2.Write( numTracks, sequence->ticksPerBeat() ) )
		{
			fprintf( stderr, "Error writing midi file\n");
			return false;
		}
		
		sequence->firstMeasure = firstMeasureValue; 
		
		return true;
		
}

bool makeJDKMidiSequence(Sequence* sequence, jdkmidi::MIDIMultiTrack& tracks, bool selectionOnly,
						 /*out*/int* songLengthInTicks, /*out*/int* startTick, /*out*/ int* numTracks)
{
		
		int trackLength=-1;
		int channel=0;
		
		int substract_ticks;
		
		if(selectionOnly)
		{
			//  ---------------------------- add events to tracks --------------------
			trackLength = sequence->getCurrentTrack()->addMidiEvents( tracks.GetTrack(sequence->getCurrentTrackID()+1),
																	  channel,
																	  sequence->firstMeasure,
																	  true,
																	  *startTick );
			
			substract_ticks = *startTick;
			
			if(trackLength == -1) return -1; // nothing to play in track (empty track - play nothing)
			*songLengthInTicks=trackLength;
			
		}
		else
			// play from beginning
		{
			
			(*startTick) = -1;
			
			const int trackAmount = sequence->getTrackAmount();
			for(int n=0; n<trackAmount; n++)
			{
				
				int trackFirstNote=-1;
				trackLength = sequence->getTrack(n)->addMidiEvents( tracks.GetTrack(n+1), channel, sequence->firstMeasure, false, trackFirstNote );
				
				if( (trackFirstNote<(*startTick) and trackFirstNote != -1) or (*startTick)==-1) (*startTick) = trackFirstNote;
				
				
				if(trackLength == -1) continue; // nothing to play in track (empty track - skip it)
				if(trackLength > *songLengthInTicks) *songLengthInTicks=trackLength;
				
				channel++; if (channel==9) channel++;
				
			}
			substract_ticks = *startTick;
			
		}
		
		
		if(*songLengthInTicks < 1) return -1; // nothing to play at all (empty song - play nothing)
		*numTracks = sequence->getTrackAmount()+1;
		
		// --------------------------------- time sig --------------------
		{
			jdkmidi::MIDITimedBigMessage m;
			m.SetTime( 0 );
			
			float denom = (float)log(sequence->getMeasureBottomValue())/(float)log(2);
			m.SetTimeSig( sequence->getMeasureTopValue(), (int)denom );  
			
			if( !tracks.GetTrack(0)->PutEvent( m ) ) { std::cout << "Error adding event" << std::endl;  return false; }
		}
		
		
		// --------------------------------- tempo --------------------
		
		jdkmidi::MIDITimedBigMessage m;
		m.SetTime( 0 );
		m.SetTempo32( sequence->getTempo() * 32 ); // tempo stored as bpm * 32, giving 1/32 bpm resolution
		
		if( !tracks.GetTrack(0)->PutEvent( m ) ) { std::cout << "Error adding event" << std::endl;  return false; }
		
		for(int n=0; n<sequence->tempoEvents.size(); n++)
		{
			
			jdkmidi::MIDITimedBigMessage m;
			
			int tempo_time = sequence->tempoEvents[n].getTick() - substract_ticks;
			
			if(tempo_time < 0)
			{
				if(n+1<sequence->tempoEvents.size()) continue; // event is not in played measures, ignore it
				else tempo_time = 0; // we need to consider event because it's the last and will affect future measures
			}
			
			m.SetTime( tempo_time );
			m.SetTempo32( convertTempoBendToBPM(sequence->tempoEvents[n].getValue() ) * 32 /* tempo is stored as bpm * 32, giving 1/32 bpm resolution*/ );
			
			if( !tracks.GetTrack(0)->PutEvent( m ) ) { std::cout << "Error adding event" << std::endl;  return false; }
			
		}
		
		// --------------------------------- copyright ---------------------------------
		if(sequence->getCopyright().size()>0)
		{
			
			jdkmidi::MIDITimedBigMessage m;
			
			m.SetText( 2 );
			m.SetByte1( 2 );
			
			jdkmidi::MIDISystemExclusive sysex( (unsigned char*)sequence->getCopyright().c_str(),
												sequence->getCopyright().size()+1,
												sequence->getCopyright().size()+1, false);
			
			m.CopySysEx( &sysex );
			
			m.SetTime( 0 );
			
			if( !tracks.GetTrack(0)->PutEvent( m ) ) { std::cout << "Error adding event" << std::endl; return false; }
			
			
		}
		
		// ---------------------------- add dummy event after the actual end to ensure it doesn't stop playing too quickly --------------------------
		// adds event way after actual stop point, to make sure song the midi player will reach the last actual note before stopping
		// (e.g. i had issues with Quicktime stopping playback too soon and never actually reaching end of song event)
		
		{
			jdkmidi::MIDITimedBigMessage m;
			m.SetTime( *songLengthInTicks + sequence->ticksPerBeat()*3 );
			m.SetControlChange( 0,
								127,
								0 );
			
			for(int n=0; n<sequence->getTrackAmount(); n++)
			{
				if( !tracks.GetTrack(n+1)->PutEvent( m ) ) { std::cout << "Error adding midi event!" << std::endl; }
			}//next
		}
		
		return true;
}

void makeMidiBytes(Sequence* sequence, bool selectionOnly, /*out*/int* songlength, /*out*/int* startTick, /*out*/char** midiSongData, /*out*/int* datalength)
{
	int numTracks = -1;
	
	jdkmidi::MIDIMultiTrack tracks;
		
	makeJDKMidiSequence(sequence, tracks, selectionOnly, songlength, startTick, &numTracks);
	
	// create the output stream
	MidiToMemoryStream* out_stream=new MidiToMemoryStream();
	
	jdkmidi::MIDIFileWriteMultiTrack writer(
											&tracks,
											out_stream
											);
	
	// write the output data
	if( !writer.Write( numTracks, sequence->ticksPerBeat() ) )
	{
		fprintf( stderr, "Error writing midi file\n");
		return;
	}
	
	(*midiSongData) = (char*)malloc(out_stream->getDataLength());
	
	out_stream->storeMidiData( (*midiSongData) );
	
	*datalength = out_stream->getDataLength();
	
	delete out_stream;
}

int convertTempoBendToBPM(int val)
{
	return (int)( (127-val)*380.0/128.0 + 20);
}
/*

bool exportMidiFile(Sequence* sequence, char* filepath)
{
	// when we're saving, we always want song to start at first measure, so temporarly switch firstMeasure to 0, and set it back in the end
	const int firstMeasureValue=sequence->firstMeasure;
    sequence->firstMeasure=0;   
		
	jdkmidi::MIDIMultiTrack tracks;
	int length = -1, start = -1, numTracks = -1;
	makeJDKMidiSequence(sequence, tracks, false, &length, &start, &numTracks);
	
	jdkmidi::MIDIFileWriteStreamFileName file_stream( filepath );
	
	jdkmidi::MIDIFileWriteMultiTrack writer2(
											 &tracks,
											 &file_stream
											 );
	
	// write the output file
	if( !writer2.Write( numTracks, sequence->ticksPerBeat() ) )
	{
		fprintf( stderr, "Error writing midi file\n");
		return false;
	}
	
	sequence->firstMeasure = firstMeasureValue; 
	
	return true;
	
}

bool playSequence(Sequence* sequence,  int* startTick)
{
	char* data;
	int datalength = -1;
	
	int songLengthInTicks = -1;
	makeMidiBytes(sequence, false, &songLengthInTicks, startTick, &data, &datalength);
	
	sequence->glPane->getMusicPlayer()->playMidiData( data, datalength, songLengthInTicks + sequence->ticksPerBeat() );
	
	free(data);
	
	return true;
}

bool playSelected(Sequence* sequence,  int* startTick)
{
	char* data;
	int datalength = -1;
	int songLengthInTicks = -1;

	makeMidiBytes(sequence, true, &songLengthInTicks, startTick, &data, &datalength);
	
	sequence->glPane->getMusicPlayer()->playMidiData( data, datalength, songLengthInTicks + sequence->ticksPerBeat() );
	
	free(data);
	
	return true;
}

void exportAudioFile(Sequence* sequence, char* filepath)
{
	// when we're saving, we always want song to start at first measure, so temporarly switch firstMeasure to 0, and set it back in the end
	const int firstMeasureValue=sequence->firstMeasure;
    sequence->firstMeasure=0; 
	
	char* data;
	int length = -1;
	
	int startTick = -1, songlength = -1;
	makeMidiBytes(sequence, false, &songlength, &startTick, &data, &length);
	
	PlatformMidiManager::exportToAudio( data, length, filepath );
	
	free(data);
	
	sequence->firstMeasure = firstMeasureValue; 
}

*/
/*
 * This function converts the set of object-oriented instances Aria uses to represent music into midi bytes, using libjdkmidi
 */
/*PLAY, PLAYSELECTED, SAVE_MIDI or SAVE_SAMPLED_AUDIO*/
/*
int makeMidiData(Sequence* sequence, int mode , char* filepath)
{
	
	jdkmidi::MIDIMultiTrack tracks;
	makeJDKMidiSequence(sequence, tracks);
    
	// find out what to do
	bool save_to_memory = false, save_to_file=false;
	
	if( ((mode==PLAY or mode==PLAYSELECTED) and PlatformMidiManager::canPlayMidiBytes()) or
		(mode==SAVE_SAMPLED_AUDIO and PlatformMidiManager::savesToSampledFromBytes()) ) save_to_memory = true;
	else save_to_file=true;
	
	// ------------------- throw data to a file (used when saving as midi) -------------
	if(save_to_file)
	{
		
		jdkmidi::MIDIFileWriteStreamFileName file_stream( filepath );
		
		jdkmidi::MIDIFileWriteMultiTrack writer2(
												 &tracks,
												 &file_stream
												 );
		
		// write the output file
		if( !writer2.Write( num_tracks, sequence->ticksPerBeat() ) )
		{
			fprintf( stderr, "Error writing midi file\n");
			return false;
		}
		
		if(mode == PLAY or mode == PLAYSELECTED)
		{
			sequence->glPane->getMusicPlayer()->playMidiData( "aria_temp.mid", songLengthInTicks + sequence->ticksPerBeat() );
			return startTick;
		}
		
	}
	
	
	// --------------------- load data into memory ---------------------
    if(save_to_memory)
    {
		
		// ---------------------------- add dummy event after the actual end to ensure it doesn't stop playing too quickly --------------------------
		// adds event way after actual stop point, to make sure song the midi player will reach the last actual note before stopping
		// (e.g. i had issues with Quicktime stopping playback too soon and never actually reaching end of song event)
		
	{
		jdkmidi::MIDITimedBigMessage m;
		m.SetTime( songLengthInTicks + sequence->ticksPerBeat()*3 );
		m.SetControlChange( 0,
							127,
							0 );
		
		if(mode==PLAYSELECTED)
		{
			if( !tracks.GetTrack(sequence->currentTrack+1)->PutEvent( m ) ) { std::cout << "Error adding midi event!" << std::endl; }
		}
		else
		{
			
			for(int n=0; n<sequence->getTrackAmount(); n++)
			{
				if( !tracks.GetTrack(n+1)->PutEvent( m ) ) { std::cout << "Error adding midi event!" << std::endl; }
			}//next
			
		}//endif
	}
		
		
        // create the output stream
        MidiToMemoryStream* out_stream=new MidiToMemoryStream();
        
        jdkmidi::MIDIFileWriteMultiTrack writer(
                                                &tracks,
                                                out_stream
                                                );
        
        // write the output data
        if( !writer.Write( num_tracks, sequence->ticksPerBeat() ) )
		{
            fprintf( stderr, "Error writing midi file\n");
            return -1;
        }
        
        char midiSongData[out_stream->getDataLength()];
        
        out_stream->storeMidiData(midiSongData);
        
        if(mode == PLAY or mode == PLAYSELECTED)
            sequence->glPane->getMusicPlayer()->playMidiData( midiSongData, out_stream->getDataLength(), songLengthInTicks + sequence->ticksPerBeat() );
		
        if(mode == SAVE_SAMPLED_AUDIO)
            PlatformMidiManager::exportToAudio( midiSongData, out_stream->getDataLength(), filepath );
        
        delete out_stream;
    }
	
	
    // set back correct first measure value now that exporting is over
    if(mode==SAVE_MIDI or mode==SAVE_SAMPLED_AUDIO)
	{
        sequence->firstMeasure=firstMeasureValue;   
    }
    
    return startTick;
    
}
*/

}
