/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef _track_h_
#define _track_h_

#include "irrXML/irrXML.h"

#include "wx/wfstream.h"

#include "jdkmidi/world.h"
#include "jdkmidi/track.h"
#include "jdkmidi/multitrack.h"
#include "jdkmidi/filereadmultitrack.h"
#include "jdkmidi/fileread.h"
#include "jdkmidi/fileshow.h"
#include "jdkmidi/filewritemultitrack.h"
#include "jdkmidi/msg.h"
#include "jdkmidi/sysex.h"

#include "jdkmidi/sequencer.h"
#include "jdkmidi/driver.h"
#include "jdkmidi/process.h"

#include <vector>
#include <string>

#include "Midi/Note.h"
#include "Midi/ControllerEvent.h"

#include "ptr_vector.h"

namespace AriaMaestosa {
	
class Sequence; // forward
class GraphicalTrack;
class MainFrame;
class ControllerEvent;
class GLPane;



const int SELECTED_NOTES = -1;
const int ALL_NOTES = -2;

class Track {
	
	DECLARE_LEAK_CHECK();
	
    MainFrame* frame;
    int trackUniqueID;
    ptr_vector<Note> notes;
	ptr_vector<Note> noteOff;
    ptr_vector<ControllerEvent> controllerEvents;
    int trackid;
    
	std::string name;
	
public:
        
    
    // ------------- read-only -------------
    GraphicalTrack* graphics;
    Sequence* sequence;
    GLPane* glPane;
    // -------------------------------------
    
    Track(MainFrame* parent, Sequence* sequence);
    ~Track();
	
	// replace events in time order
	void reorderNoteVector();
	void reorderNoteOffVector();
    
    void setId(const int id);
    
	// undo/redo
    void loadUndoMemory();
    void saveUndoMemory();
    
    int getGridDivider();
    
	void setName(char* name);
	std::string& getName();
	
	void reserveMemoryForEvents(int eventAmount);
		
    // get info on notes
    int getNoteAmount();
    int getNoteStartInPixels(const int id);
    int getNoteEndInPixels(const int id);
    int getNoteStartInMidiTicks(const int id);
    int getNoteEndInMidiTicks(const int id);
    int getNotePitchID(const int id);
    bool isNoteSelected(const int id);
    int getNoteVolume(const int id);
    
	void playNote(const int id, const bool noteChange=false);
	
    int getNoteString(const int id); // guitar editor
    int getNoteFret(const int id); // guitar editor
    void prepareNotesForGuitarEditor();
    
    // returns the amount of ALL types of controller, not only of specified type
    // the only goal of id is to determine whether the app is searching for a control event or for a tempo event
    int getControllerEventAmount(const int controllerTypeID);
    ControllerEvent* getControllerEvent(const int id, const int controllerTypeID);
    
    bool addNote(const int pitchID, const int startTick, const int endTick, const int volume, const int string=-1);
    bool addNote( Note* note, bool check_for_overlapping_notes=true );
    void addController(const int x, const int value, const int controller);
    void addControllerSlide(const int x1, const int value1, const int x2, const int value2, const int controller);
    
    // this one should only be called by the midi file loader.
    // its only difference is that it knows all events are OK and that they are in time order, so it doesn't waste time doing checks
    void addControllerFromMidiFileReader(const int x, const int value, const int controller);
    
    // modify notes
    void moveNotes(const int relativeX, const int relativeY, const int noteID); 
    void setNoteVolume(const int volume, const int noteID);
    void setNoteEnd(const int tick, const int noteID);
    void resizeNotes(const int relativeWidth, const int noteID);
    void selectNote(const int id, const bool selected, bool ignoreModifiers=false);
    void deleteSelected();
    void removeOverlapping();
	void removeNote(const int id);
		
	void insertEmptyMeasures(int measureID, int amount);
	void removeMeasures(int from_measure, int to_measure);
    void snapNotesToGrid();
    
    int getFirstNoteTick(bool selectionOnly=false);
    
    void scale(float factor, int relative_to, bool selectionOnly);
    
    // guitar edtior
    void shiftFrets(const int amount, const int id);
    void shiftString(const int amount, const int id);
    void numberPressed(int number);    
    void updateNoteTuning(const int id);
    
    void copy();
    void paste(bool atMouse);
    
    // playback/file IO
    int addMidiEvents(jdkmidi::MIDITrack* track, int channel, int firstMeasure, bool selectionOnly, int& startTick); // returns length
    
    // serialization
    void saveToFile(wxFileOutputStream& fileout);
    bool readFromFile(irr::io::IrrXMLReader* xml);
};

}

#endif
